/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: policygraph.cc                               */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#include "policygraph.h"

// *********************************************
// ************** BGP Policy graph *************
// *********************************************

/* Warning: using this code assumes that your compiler supports partial
   expression evaluation (e.g., in (A || B), if A is true, B is not
   evaluated at all) */


/******* Method implementations for the Date class follow ********/

// Default constructor (sets date to Jan 01, 1900)
Date::Date() {
   year = 1900;
   month = 1;
   day = 1;
}

// Copy constructor
Date::Date(const Date& otherDate) {
   year = otherDate.year;
   month = otherDate.month;
   day = otherDate.day;
}

// This method initializes a date object with given day, month, year
Date::Date(char theDay, char theMonth, int theYear) {
   year = theYear;
   month = theMonth;
   day = theDay;
}

/* This constructor builds a Date object starting from a string in the
   format YYYYMMDD */
Date::Date(char* date) {
   fromString(date);
}

// Assignment operator
Date& Date::operator= (const Date& otherDate) {
   year = otherDate.year;
   month = otherDate.month;
   day = otherDate.day;
   
   return *this;
}

// Equality test
bool Date::operator== (const Date& otherDate) {
   return (year == otherDate.year &&
           month == otherDate.month &&
           day == otherDate.day);
}

// Inequality test
bool Date::operator!= (const Date& otherDate) {
   return !(*this == otherDate);
}

// Less than operator
bool Date::operator< (const Date& otherDate) {
   return (year < otherDate.year ||
           (year == otherDate.year && month < otherDate.month) ||
           (year == otherDate.year && month == otherDate.month && day < otherDate.day));
}

// Less than or equal to operator
bool Date::operator<= (const Date& otherDate) {
   return (*this < otherDate || *this == otherDate);
}

// Greater than operator
bool Date::operator> (const Date& otherDate) {
   return (*this != otherDate && !(*this < otherDate));
}

// Greater than or equal to operator
bool Date::operator>= (const Date& otherDate) {
   return !(*this < otherDate);
}

// Builds a Date object starting from a string in the format YYYYMMDD
void Date::fromString(char* date) {
   long int value;
   char *strangeCharacter;

   if (strlen(date) != 8) return;          // Malformed date
   
   value = strtol(strndup(date, 4), &strangeCharacter, 10);
   if (*strangeCharacter != '\0') return;   // Malformed date
   year = value;
   date += 4;
   
   value = strtol(strndup(date, 2), &strangeCharacter, 10);
   if (*strangeCharacter != '\0') return;   // Malformed date
   month = value;
   date += 2;
   
   value = strtol(strndup(date, 2), &strangeCharacter, 10);
   if (*strangeCharacter != '\0') return;   // Malformed date
   day = value;
}




/******* Method implementations for the Policy class follow ********/

/* A simple support method that releases memory (if it had been
   allocated before) */
void Policy::release_mem(char **mem_zone) {
   if (*mem_zone != NULL) {
      delete [] *mem_zone;
      *mem_zone = NULL;
   }
}

/* This structure is only used to support automatic memory
   allocation/release for edge labels */

// Default constructor
Policy::Policy() {
   import_policy = new char;
   import_policy[0] = '\0';          // Setup an empty policy
   import_policy_registry = new char;
   import_policy_registry[0] = '\0'; // Setup an empty policy registry
   export_policy = new char;
   export_policy[0] = '\0';          // Setup an empty policy
   export_policy_registry = new char;
   export_policy_registry[0] = '\0'; // Setup an empty policy registry
   oriented = false;
}

// Copy constructor
Policy::Policy(const Policy &otherPolicy) {
   oriented = otherPolicy.oriented;
   import_policy = new char[strlen(otherPolicy.import_policy) + 1]; // One more character for '\0'
   strcpy (import_policy, otherPolicy.import_policy);
   import_policy_registry = new char[strlen(otherPolicy.import_policy_registry) + 1]; // One more character for '\0'
   strcpy (import_policy_registry, otherPolicy.import_policy_registry);
   export_policy = new char[strlen(otherPolicy.export_policy) + 1]; // One more character for '\0'
   strcpy (export_policy, otherPolicy.export_policy);
   export_policy_registry = new char[strlen(otherPolicy.export_policy_registry) + 1]; // One more character for '\0'
   strcpy (export_policy_registry, otherPolicy.export_policy_registry);
   import_policy_date = otherPolicy.import_policy_date;
   export_policy_date = otherPolicy.export_policy_date;
}

// Destructor
Policy::~Policy() {
   Policy::release_mem (&import_policy);
   Policy::release_mem (&import_policy_registry);
   Policy::release_mem (&export_policy);
   Policy::release_mem (&export_policy_registry);
}

// Assignment operator
Policy& Policy::operator=(const Policy &otherPolicy) {
   release_mem (&import_policy);
   Policy::release_mem (&import_policy_registry);
   release_mem (&export_policy);
   Policy::release_mem (&export_policy_registry);
   import_policy = new char[strlen(otherPolicy.import_policy) + 1]; // One more character for \0
   strcpy (import_policy, otherPolicy.import_policy);
   import_policy_registry = new char[strlen(otherPolicy.import_policy_registry) + 1]; // One more character for \0
   strcpy (import_policy_registry, otherPolicy.import_policy_registry);
   export_policy = new char[strlen(otherPolicy.export_policy) + 1]; // One more character for \0
   strcpy (export_policy, otherPolicy.export_policy);
   export_policy_registry = new char[strlen(otherPolicy.export_policy_registry) + 1]; // One more character for \0
   strcpy (export_policy_registry, otherPolicy.export_policy_registry);
   oriented = otherPolicy.oriented;
   import_policy_date = otherPolicy.import_policy_date;
   export_policy_date = otherPolicy.export_policy_date;
   return (*this);
}


/* This method can be used to set the import/export policy of an
   AS. Considering an edge AS1-AS2, this label represents
   the import/export policy of AS1 */
void Policy::setPolicy(policy_type pol, char* policy) {
   switch (pol) {
      case POL_IMPORT:
         release_mem (&import_policy);
         import_policy = new char[strlen(policy) + 1]; // One more character for '\0'
         strcpy (import_policy, policy);
         break;
      case POL_EXPORT:
         release_mem (&export_policy);
         export_policy = new char[strlen(policy) + 1]; // One more character for '\0'
         strcpy (export_policy, policy);
         break;
      default:
         // User must have made some mistake: forgive him by doing nothing
         break;
   }
}

/* This method is similar to the above setPolicy, but it also allows to store
   a date (last update of a policy) and a registry the policy has been
   found in. */
void Policy::setPolicy(policy_type pol, char* policy, Date update, char* registry) {
   switch (pol) {
      case POL_IMPORT:
         release_mem (&import_policy);
         release_mem (&import_policy_registry);
         import_policy = new char[strlen(policy) + 1]; // One more character for '\0'
         strcpy (import_policy, policy);
         import_policy_registry = new char[strlen(registry) + 1]; // One more character for '\0'
         strcpy (import_policy_registry, registry);
         import_policy_date = update;
         break;
      case POL_EXPORT:
         release_mem (&export_policy);
         release_mem (&export_policy_registry);
         export_policy = new char[strlen(policy) + 1]; // One more character for '\0'
         strcpy (export_policy, policy);
         export_policy_registry = new char[strlen(registry) + 1]; // One more character for '\0'
         strcpy (export_policy_registry, registry);
         export_policy_date = update;
         break;
      default:
         // User must have made some mistake: forgive him by doing nothing
         break;
   }   
}

/* This method updates date and registry information for a given import/export
   policy. */
void Policy::setPolicy(policy_type pol, Date update, char* registry) {
   switch (pol) {
      case POL_IMPORT:
         release_mem (&import_policy_registry);
         import_policy_registry = new char[strlen(registry) + 1]; // One more character for '\0'
         strcpy (import_policy_registry, registry);
         import_policy_date = update;
         break;
      case POL_EXPORT:
         release_mem (&export_policy_registry);
         export_policy_registry = new char[strlen(registry) + 1]; // One more character for '\0'
         strcpy (export_policy_registry, registry);
         export_policy_date = update;
         break;
      default:
         // User must have made some mistake: forgive him by doing nothing
         break;
   }
}

// This method updates date information for a given import/export policy.
void Policy::setPolicy(policy_type pol, Date update) {
   switch (pol) {
      case POL_IMPORT:
         import_policy_date = update;
         break;
      case POL_EXPORT:
         export_policy_date = update;
         break;
      default:
         // User must have made some mistake: forgive him by doing nothing
         break;
   }
}

/* This method returns the current import/export policy of an AS,
   if any (returns an empty string if a policy has not been set yet).
   Given an edge AS1-AS2, the method returns the import/export
   policy of AS1. */
char* Policy::getPolicy(policy_type pol) {
   switch (pol) {
      case POL_IMPORT:
         return import_policy;
         break;
      case POL_EXPORT:
         return export_policy;
         break;
      default:
         // Absent-minded user: policy type is invalid, hence we do nothing
         break;
   }
   // Dummy return (just for suppressing compiler warnings)
   return NULL;
}

// This method returns the date associated with an import/export policy
Date Policy::getPolicyDate(policy_type pol) {
   switch (pol) {
      case POL_IMPORT:
         return import_policy_date;
         break;
      case POL_EXPORT:
         return export_policy_date;
         break;
      default:
         // What's this policy about? Maybe user was thinking about something else...
         break;
   }
   // Dummy return
   return Date();
}

// This method returns the registry a given policy has been found in.
char* Policy::getPolicyRegistry(policy_type pol) {
   switch (pol) {
      case POL_IMPORT:
         return import_policy_registry;
         break;
      case POL_EXPORT:
         return export_policy_registry;
         break;
      default:
         // What's this policy about? Maybe user was thinking about something else...
         break;
   }
   // Dummy return
   return NULL;
}




/******* Method implementations for the PolicyGraph class follow ********/

/* The following constant stores the 3-characters tag that identifies
   policy graphs when being written to file */
const char PolicyGraph::policyGraphMarker[4] = "TPG";

// Default constructor
PolicyGraph::PolicyGraph() {
   enhanced = false;
   strcpy (error, "");
}

// Destructor
PolicyGraph::~PolicyGraph() {
}


/* This funcion converts an edge direction (IN, OUT, NONE)
   into a printable integer value (-1, 1, 0) */
int PolicyGraph::dirToVal(edge_dir dir) {
   switch (dir) {
      case DIR_IN:   return -1;
                     break;
      case DIR_OUT:  return 1;
                     break;
      case DIR_NONE: return 0;
                     break;
      default:       return 0;
                     break;
   }
}

/* This function converts an integer value (-1, 1, 0)
   into an edge direction (IN, OUT, NONE) */
edge_dir PolicyGraph::valToDir(int val) {
   switch(val) {
      case -1: return DIR_IN;
               break;
      case 1:  return DIR_OUT;
               break;
      case 0:  return DIR_NONE;
               break;
      default: return DIR_NONE;
               break;
   }
}



/* Dumps all the policy graph contents to an
   ASCII file; use `readFromFile' to
   get them back; the only parameter is
   a pointer to an open file strem;
   this method returns true if operation
   is successful; false otherwise */
bool PolicyGraph::writeToFile(FILE *f) {
   List <graph_node*> nodeList;
   List <graph_edge*> edgeList;
   graph_node *curNode;
   graph_edge *curEdge;
   bool stop, goOn;
   
   strcpy (error, "");
   // Write the string that identifies the graph
   if (!fprintf (f, "%s\n", policyGraphMarker)) {
      strcat (error, "Could not write file type string");
      return false;
   }
   
   // Write number of nodes
   if (!fprintf (f, "%u\n", nodes())) {
      strcat (error, "Could not write number of nodes");
      return false;
   }

   // Write number of edges
   if (!fprintf (f, "%u\n", edges())) {
      strcat (error, "Could not write number of edges");
      return false;
   }

   // Write list of nodes
   nodeList = listNodes();
   stop = false;
   goOn = !nodeList.isEmpty();
   for (nodeList.goToFirst(); goOn; goOn = nodeList.next()) {
      curNode = nodeList.current();
      if (!stop && !fprintf (f, "%i\n", *(curNode->contents))) {
         strcat (error, "Could not write node");
         sprintf(error, "%s %i", error, *(curNode->contents));
         /* There is no immediate return, because that would not
            allow memory deallocation */
         stop = true;
      }
      delete curNode;
   }
   
   if (stop) return false;
   
   // Write list of edges
   edgeList = listEdges();
   goOn = !edgeList.isEmpty();
   for (edgeList.goToFirst(); goOn; goOn = edgeList.next()) {
      curEdge = edgeList.current();
      if (!stop && !fprintf (f, "%i %i %i %i %i %s %i %s %4i %2i %2i %i %s %i %s %4i %2i %2i %i %s %i %s %4i %2i %2i %i %s %i %s %4i %2i %2i\n",
            *(curEdge->from->contents),
            *(curEdge->to->contents),
            (curEdge->weight_from->oriented ? 1 : 0),
            dirToVal(*(curEdge->orientation)),
            
            strlen(curEdge->weight_from->getPolicy(POL_IMPORT)),
            curEdge->weight_from->getPolicy(POL_IMPORT),
            strlen(curEdge->weight_from->getPolicyRegistry(POL_IMPORT)),
            curEdge->weight_from->getPolicyRegistry(POL_IMPORT),
            curEdge->weight_from->getPolicyDate(POL_IMPORT).year,
            curEdge->weight_from->getPolicyDate(POL_IMPORT).month,
            curEdge->weight_from->getPolicyDate(POL_IMPORT).day,
            
            strlen(curEdge->weight_from->getPolicy(POL_EXPORT)),
            curEdge->weight_from->getPolicy(POL_EXPORT),
            strlen(curEdge->weight_from->getPolicyRegistry(POL_EXPORT)),
            curEdge->weight_from->getPolicyRegistry(POL_EXPORT),
            curEdge->weight_from->getPolicyDate(POL_EXPORT).year,
            curEdge->weight_from->getPolicyDate(POL_EXPORT).month,
            curEdge->weight_from->getPolicyDate(POL_EXPORT).day,
            
            strlen(curEdge->weight_to->getPolicy(POL_IMPORT)),
            curEdge->weight_to->getPolicy(POL_IMPORT),
            strlen(curEdge->weight_to->getPolicyRegistry(POL_IMPORT)),
            curEdge->weight_to->getPolicyRegistry(POL_IMPORT),
            curEdge->weight_to->getPolicyDate(POL_IMPORT).year,
            curEdge->weight_to->getPolicyDate(POL_IMPORT).month,
            curEdge->weight_to->getPolicyDate(POL_IMPORT).day,
            
            
            strlen(curEdge->weight_to->getPolicy(POL_EXPORT)),
            curEdge->weight_to->getPolicy(POL_EXPORT),
            strlen(curEdge->weight_to->getPolicyRegistry(POL_EXPORT)),
            curEdge->weight_to->getPolicyRegistry(POL_EXPORT),
            curEdge->weight_to->getPolicyDate(POL_EXPORT).year,
            curEdge->weight_to->getPolicyDate(POL_EXPORT).month,
            curEdge->weight_to->getPolicyDate(POL_EXPORT).day
            )) {
         strcat (error, "Could not write edge");
         sprintf(error, "%s %i %i", error, *(curEdge->from->contents),
                                          *(curEdge->to->contents));
         /* There is no immediate return, because that would not
            allow memory deallocation */
         stop = true;
      }

      delete curEdge->from;
      delete curEdge->to;
      delete curEdge;
   }

   return true;
}

/* Builds a policy graph from the data that have
   previously been generated by `writeToFile';
   the only parameter is a pointer to an
   open file stream; returns true if
   operation is successful; false otherwise */
bool PolicyGraph::readFromFile(FILE *f) {
   char string[MAX_LINE_LENGTH + 1], oldString[MAX_LINE_LENGTH + 1];             // One more byte for '\0'
   char *err, *curToken, *registry;
   int intValue;
   unsigned int uValue, nodeCnt, edgeCnt;
   unsigned long int ulValue, policyLength;
   ASnode from, to;
   Policy pol1, pol2;
   Date policyDate;
   edge_dir direction;


   strcpy (error, "");

   // Get graph type
   if (!fgets(string, MAX_LINE_LENGTH, f)) {
      strcat (error, "Could not retrieve file type string");
      return false;
   }
   
   // Remove CR/LF at end of line
   if (strpbrk(string, "\n\r")) {
      *(strpbrk(string, "\n\r")) = '\0';
   }
   if (strcmp(string, policyGraphMarker) != 0) {
      strcat (error, "Not a policy graph");
      return false;
   }

   // Get number of nodes
   if (!fgets(string, MAX_LINE_LENGTH, f)) {
      strcat (error, "Could not retrieve number of nodes");
      return false;
   }
   // Remove CR/LF at end of line
   if (strpbrk(string, "\n\r")) {
      *(strpbrk(string, "\n\r")) = '\0';
   }
   uValue = strtoul(string, &err, 10);
   if (strlen(err) > 0) {
      strcat (error, "Invalid number of nodes: `");
      strcat (error, string);
      strcat (error, "'");
      return false;
   }
   nodeCnt = uValue;
   
   // Get number of edges
   if (!fgets(string, MAX_LINE_LENGTH, f)) {
      strcat (error, "Could not retrieve number of edges");
      return false;
   }
   // Remove CR/LF at end of line
   if (strpbrk(string, "\n\r")) {
      *(strpbrk(string, "\n\r")) = '\0';
   }
   uValue = strtoul(string, &err, 10);
   if (strlen(err) > 0) {
      strcat (error, "Invalid number of edges: `");
      strcat (error, string);
      strcat (error, "'");
      return false;
   }
   edgeCnt = uValue;

   // Get nodes
   for (unsigned int i = 0; nodeCnt > 0 && i <= nodeCnt - 1; i++) {
      if (!fgets(string, MAX_LINE_LENGTH, f)) {
         strcat (error, "Unable to retrieve all the nodes");
         return false;
      }
      // Remove CR/LF at end of line
      if (strpbrk(string, "\n\r")) {
         *(strpbrk(string, "\n\r")) = '\0';
      }
      intValue = strtol(string, &err, 10);
      if (intValue < 1 || intValue > 65535 || (strlen(err) > 0)) {
         strcat (error, "Invalid AS number: `");
         strcat (error, string);
         strcat (error, "'");
         return false;
      }
      addNode((ASnode) intValue, (ASnode) intValue);
   }
   
   // Get edges
   for (unsigned int i = 0; edgeCnt > 0 && i <= edgeCnt - 1; i++) {
      if (!fgets(string, MAX_LINE_LENGTH, f)) {
         strcat (error, "Unable to retrieve all the edges");
         return false;
      }
      // Remove CR/LF at end of line
      if (strpbrk(string, "\n\r")) {
         *(strpbrk(string, "\n\r")) = '\0';
      }
      /* String will be tokenized, so a copy
         of it should be made */
      strcpy(oldString, string);

      // First node
      curToken = strtok(string, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      intValue = strtol(curToken, &err, 10);
      if (intValue < 1 || intValue > 65535 || (strlen(err) > 0)) {
         strcat (error, "Invalid AS number: `");
         strcat (error, curToken);
         strcat (error, "'");
         return false;
      }
      from = intValue;

      // Second node
      curToken = strtok(NULL, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      intValue = strtol(curToken, &err, 10);
      if (intValue < 1 || intValue > 65535 || (strlen(err) > 0)) {
         strcat (error, "Invalid AS number: `");
         strcat (error, curToken);
         strcat (error, "'");
         return false;
      }
      to = intValue;

      // Edge direction flag
      curToken = strtok(NULL, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      if (strcmp(curToken, "0") != 0 && strcmp(curToken, "1") != 0) {
         strcat (error, "Invalid edge direction flag: `");
         strcat (error, curToken);
         strcat (error, "'");
         return false;
      }
      pol1.oriented = pol2.oriented =
         (strcmp(curToken, "1") == 0);

      // Edge direction
      curToken = strtok(NULL, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      intValue = strtol(curToken, &err, 10);
      if (intValue < -1 || intValue > 1 || (strlen(err) > 0)) {
         strcat (error, "Invalid edge direction: `");
         strcat (error, curToken);
         strcat (error, "'");
         return false;
      }
      direction = valToDir(intValue);

      /* The following comments assume these conventions:
         IMPORT-POLICY-FROM:          import policy from "this" side of the edge
                                      (i.e., given an edge AS1-AS2, this is the
                                      import policy of AS1)
         IMPORT-POLICY-FROM-REGISTRY: registry associated with IMPORT-POLICY-FROM
         IMPORT-POLICY-FROM-DATE:     latest update for IMPORT-POLICY-FROM
         EXPORT-POLICY-FROM:          export policy from "this" side of the edge
                                      (i.e., given an edge AS1-AS2, this is the
                                      export policy of AS1)
         EXPORT-POLICY-FROM-REGISTRY: registry associated with EXPORT-POLICY-FROM
         EXPORT-POLICY-FROM-DATE:     latest update for EXPORT-POLICY-FROM
         IMPORT-POLICY-TO:            import policy from "the other" side of the edge
                                      (i.e., given an edge AS1-AS2, this is the
                                      import policy of AS2)
         IMPORT-POLICY-TO-REGISTRY:   registry associated with IMPORT-POLICY-TO
         IMPORT-POLICY-TO-DATE:       latest update for IMPORT-POLICY-TO
         EXPORT-POLICY-TO:            export policy from "the other" side of the edge
                                      (i.e., given an edge AS1-AS2, this is the
                                      export policy of AS2)
         EXPORT-POLICY-TO-REGISTRY:   registry associated with EXPORT-POLICY-TO
         EXPORT-POLICY-TO-DATE:       latest update for EXPORT-POLICY-TO */

      // Length of IMPORT-POLICY-FROM
      curToken = strtok(NULL, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      ulValue = strtoul(curToken, &err, 10);
      if (strlen(err) > 0) {
         sprintf (error, "Invalid import policy length: `%s' for AS %i from edge %i %i",
                  curToken, from, from, to);
         return false;
      }
      policyLength = ulValue;
      
      // IMPORT-POLICY-FROM
      curToken += (strlen(curToken) + 1);
      if (curToken > string + strlen(oldString)) {
         sprintf (error, "Could not get import policy for AS %i from edge %i %i",
                  from, from, to);
         return false;
      }
      curToken[policyLength] = '\0';      // Split string at the end of the current policy
      pol1.setPolicy(POL_IMPORT, curToken);
      
      // Length of IMPORT-POLICY-FROM-REGISTRY
      curToken += (strlen(curToken) + 1);
      curToken = strtok(curToken, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      ulValue = strtoul(curToken, &err, 10);
      if (strlen(err) > 0) {
         sprintf (error, "Invalid import policy registry length: `%s' for AS %i from edge %i %i",
                  curToken, from, from, to);
         return false;
      }
      policyLength = ulValue;
      
      // IMPORT-POLICY-FROM-REGISTRY
      curToken += (strlen(curToken) + 1);
      if (curToken > string + strlen(oldString)) {
         sprintf (error, "Could not get import policy registry for AS %i from edge %i %i",
                  from, from, to);
         return false;
      }
      curToken[policyLength] = '\0';      // Split string at the end of the current policy
      registry = new char[strlen(curToken) + 1];
      strcpy (registry, curToken);

      // IMPORT-POLICY-FROM-DATE
      curToken += (strlen(curToken) + 1);
      curToken = strtok(curToken, " ");
      for (int i = 1; i <= 3; i++) {
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         ulValue = strtoul(curToken, &err, 10);
         if (strlen(err) > 0) {
            sprintf (error, "Invalid import policy date: `%s' for AS %i from edge %i %i",
                     curToken, from, from, to);
            return false;
         }
         switch(i) {
            case 1:
               policyDate.year = ulValue;
               break;
            case 2:
               policyDate.month = ulValue;
               break;
            case 3:
               policyDate.day = ulValue;
               break;
         }
         curToken = strtok(NULL, " ");
      }
      pol1.setPolicy(POL_IMPORT, policyDate, registry);
      delete [] registry;
      
      // Length of EXPORT-POLICY-FROM
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      ulValue = strtoul(curToken, &err, 10);
      if (strlen(err) > 0) {
         sprintf (error, "Invalid export policy length: `%s' for AS %i from edge %i %i",
                  curToken, from, from, to);
         return false;
      }
      policyLength = ulValue;
      
      // EXPORT-POLICY-FROM
      curToken += (strlen(curToken) + 1);
      if (curToken > string + strlen(oldString)) {
         sprintf (error, "Could not get export policy for AS %i from edge %i %i",
                  from, from, to);
         return false;
      }
      curToken[policyLength] = '\0';      // Split string at the end of the current policy
      pol1.setPolicy(POL_EXPORT, curToken);
      
      // Length of EXPORT-POLICY-FROM-REGISTRY
      curToken += (strlen(curToken) + 1);
      curToken = strtok(curToken, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      ulValue = strtoul(curToken, &err, 10);
      if (strlen(err) > 0) {
         sprintf (error, "Invalid export policy registry length: `%s' for AS %i from edge %i %i",
                  curToken, from, from, to);
         return false;
      }
      policyLength = ulValue;
      
      // EXPORT-POLICY-FROM-REGISTRY
      curToken += (strlen(curToken) + 1);
      if (curToken > string + strlen(oldString)) {
         sprintf (error, "Could not get export policy registry for AS %i from edge %i %i",
                  from, from, to);
         return false;
      }
      curToken[policyLength] = '\0';      // Split string at the end of the current policy
      registry = new char[strlen(curToken) + 1];
      strcpy (registry, curToken);

      // EXPORT-POLICY-FROM-DATE
      curToken += (strlen(curToken) + 1);
      curToken = strtok(curToken, " ");
      for (int i = 1; i <= 3; i++) {
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         ulValue = strtoul(curToken, &err, 10);
         if (strlen(err) > 0) {
            sprintf (error, "Invalid export policy date: `%s' for AS %i from edge %i %i",
                     curToken, from, from, to);
            return false;
         }
         switch(i) {
            case 1:
               policyDate.year = ulValue;
               break;
            case 2:
               policyDate.month = ulValue;
               break;
            case 3:
               policyDate.day = ulValue;
               break;
         }
         curToken = strtok(NULL, " ");
      }
      pol1.setPolicy(POL_EXPORT, policyDate, registry);
      delete [] registry;
      
      // Length of IMPORT-POLICY-TO
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      ulValue = strtoul(curToken, &err, 10);
      if (strlen(err) > 0) {
         sprintf (error, "Invalid import policy length: `%s' for AS %i from edge %i %i",
                  curToken, to, from, to);
         return false;
      }
      policyLength = ulValue;
      
      // IMPORT-POLICY-TO
      curToken += (strlen(curToken) + 1);
      if (curToken > string + strlen(oldString)) {
         sprintf (error, "Could not get import policy for AS %i from edge %i %i",
                  to, from, to);
         return false;
      }
      curToken[policyLength] = '\0';      // Split string at the end of the current policy
      pol2.setPolicy(POL_IMPORT, curToken);
      
      // Length of IMPORT-POLICY-TO-REGISTRY
      curToken += (strlen(curToken) + 1);
      curToken = strtok(curToken, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      ulValue = strtoul(curToken, &err, 10);
      if (strlen(err) > 0) {
         sprintf (error, "Invalid import policy registry length: `%s' for AS %i from edge %i %i",
                  curToken, to, from, to);
         return false;
      }
      policyLength = ulValue;
      
      // IMPORT-POLICY-TO-REGISTRY
      curToken += (strlen(curToken) + 1);
      if (curToken > string + strlen(oldString)) {
         sprintf (error, "Could not get import policy registry for AS %i from edge %i %i",
                  to, from, to);
         return false;
      }
      curToken[policyLength] = '\0';      // Split string at the end of the current policy
      registry = new char[strlen(curToken) + 1];
      strcpy (registry, curToken);

      // IMPORT-POLICY-TO-DATE
      curToken += (strlen(curToken) + 1);
      curToken = strtok(curToken, " ");
      for (int i = 1; i <= 3; i++) {
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         ulValue = strtoul(curToken, &err, 10);
         if (strlen(err) > 0) {
            sprintf (error, "Invalid import policy date: `%s' for AS %i from edge %i %i",
                     curToken, to, from, to);
            return false;
         }
         switch(i) {
            case 1:
               policyDate.year = ulValue;
               break;
            case 2:
               policyDate.month = ulValue;
               break;
            case 3:
               policyDate.day = ulValue;
               break;
         }
         curToken = strtok(NULL, " ");
      }
      pol2.setPolicy(POL_IMPORT, policyDate, registry);
      delete [] registry;
      
      // Length of EXPORT-POLICY-TO
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      ulValue = strtoul(curToken, &err, 10);
      if (strlen(err) > 0) {
         sprintf (error, "Invalid export policy length: `%s' for AS %i from edge %i %i",
                  curToken, to, from, to);
         return false;
      }
      policyLength = ulValue;
      
      // EXPORT-POLICY-TO
      curToken += (strlen(curToken) + 1);
      if (curToken > string + strlen(oldString)) {
         sprintf (error, "Could not get export policy for AS %i from edge %i %i",
                  to, from, to);
         return false;
      }
      curToken[policyLength] = '\0';      // Split string at the end of the current policy
      pol2.setPolicy(POL_EXPORT, curToken);
      
      // Length of EXPORT-POLICY-TO-REGISTRY
      curToken += (strlen(curToken) + 1);
      curToken = strtok(curToken, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      ulValue = strtoul(curToken, &err, 10);
      if (strlen(err) > 0) {
         sprintf (error, "Invalid export policy registry length: `%s' for AS %i from edge %i %i",
                  curToken, to, from, to);
         return false;
      }
      policyLength = ulValue;
      
      // EXPORT-POLICY-TO-REGISTRY
      curToken += (strlen(curToken) + 1);
      if (curToken > string + strlen(oldString)) {
         sprintf (error, "Could not get export policy registry for AS %i from edge %i %i",
                  to, from, to);
         return false;
      }
      curToken[policyLength] = '\0';      // Split string at the end of the current policy
      registry = new char[strlen(curToken) + 1];
      strcpy (registry, curToken);

      // EXPORT-POLICY-TO-DATE
      curToken += (strlen(curToken) + 1);
      curToken = strtok(curToken, " ");
      for (int i = 1; i <= 3; i++) {
         if (curToken == NULL) {
            strcat (error, "Invalid edge line format: `");
            strcat (error, oldString);
            strcat (error, "'");
            return false;
         }
         ulValue = strtoul(curToken, &err, 10);
         if (strlen(err) > 0) {
            sprintf (error, "Invalid export policy date: `%s' for AS %i from edge %i %i",
                     curToken, to, from, to);
            return false;
         }
         switch(i) {
            case 1:
               policyDate.year = ulValue;
               break;
            case 2:
               policyDate.month = ulValue;
               break;
            case 3:
               policyDate.day = ulValue;
               break;
         }
         curToken = strtok(NULL, " ");
      }
      pol2.setPolicy(POL_EXPORT, policyDate, registry);
      delete [] registry;
      
      addEdge(from, to, direction, pol1, pol2);
      
   }
   
   return true;
}

/* Returns a string that describes the
   reason of last file error; this method
   should be invoked after that either one
   of `writeToFile' or `readFromFile'
   returned false */
char* PolicyGraph::errReason() {
   return error;
}

// ************** BGP Policy graph *************
