/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: policygraph.h                                */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#ifndef POLGRAPH_H
#define POLGRAPH_H

#include "graph.h"
#include <stdio.h>
#include <string.h>

#define MAX_LINE_LENGTH 4096

// *********************************************
// ************** BGP Policy Graph *************
// *********************************************

typedef unsigned short int ASnode;

// Policies are distinguished into import and export policies
enum policy_type{POL_IMPORT = 0, POL_EXPORT = 1};


/* This class is used in order to provide basic support for
   dates inside edge labels */
class Date{
   public:
      int year;
      char month;                          // January = 1
      char day;
      Date();                              /* Default constructor (sets date to
                                              Jan 01, 1900) */
      Date(const Date&);                   // Copy constructor
      Date(char theDay, char theMonth,
           int theYear);                   /* This method initializes a date
                                              object with given day, month, year */
      Date(char*);                         /* This constructor builds a Date object
                                              starting from a string in the
                                              format YYYYMMDD */
      Date& operator= (const Date&);       // Assignment operator
      bool operator== (const Date&);       // Equality test
      bool operator!= (const Date&);       // Inequality test
      bool operator< (const Date&);        // Less than operator
      bool operator<= (const Date&);       // Less than or equal to operator
      bool operator> (const Date&);        // Greater than operator
      bool operator>= (const Date&);       // Greater than or equal to operator
      void fromString(char*);              /* Builds a Date object starting from a string
                                              in the format YYYYMMDD */
};

/* This class is only used to support automatic memory
   allocation/release for edge labels */
class Policy{
   private:
      static void release_mem(char**);     /* A simple support method that releases memory
                                              (if it had been allocated before) */
      char *import_policy,                 // Label describing the import policy
           *import_policy_registry;        // Name of the registry the import policy comes from
      Date import_policy_date;             // Last update for the import policy
      char *export_policy,                 // Label describing the export policy
           *export_policy_registry;        // Name of the registry the export policy comes from
      Date export_policy_date;             // Last update for the export policy
   public:
      Policy();                            // Constructor (allocates memory)
      ~Policy();                           // Destructor  (releases allocated memory)
      Policy(const Policy&);               // Copy constructor
      Policy& operator=(const Policy&);    // Assignment operator
      bool oriented;                       /* This flag is true when the edge has been assigned
                                              a valid orientation; false otherwise */
      void setPolicy(policy_type, char*);  /* This method can be used to set the import/export policy of an
                                              AS. Considering an edge AS1-AS2, this label represents
                                              the import/export policy of AS1 */
      void setPolicy(policy_type, char*,
                     Date, char*);         /* This method is similar to the above setPolicy, but
                                              it also allows to store a date (last update
                                              of a policy) and a registry the policy has been
                                              found in. */
      void setPolicy(policy_type, Date,
                     char*);               /* This method updates date and registry information
                                              for a given import/export policy. */
      void setPolicy(policy_type, Date);   /* This method updates date information for a given import/export
                                              policy. */
      char* getPolicy(policy_type);        /* This method returns the current import/export policy of an AS,
                                              if any (returns an empty string if a policy has not been set yet).
                                              Given an edge AS1-AS2, the method returns the import
                                              policy of AS1. */
      Date getPolicyDate(policy_type);     /* This method returns the date associated with an import/export
                                              policy */
      char* getPolicyRegistry(policy_type);/* This method returns the registry a given policy has been
                                              found in. */
};

class PolicyGraph: public Graph<ASnode, Policy, ASnode> {
   private:
      char error[513];                          /* This variable stores a short
                                                   explanation for error conditions,
                                                   when they occur; 513 includes 512
                                                   characters plus terminating '\0' */
      int dirToVal(edge_dir);                   /* This funcion converts an edge direction
                                                   (IN, OUT, NONE) into a printable integer
                                                   value (-1, 1, 0) */
      edge_dir valToDir(int);                   /* This function converts an integer value
                                                   (-1, 1, 0) into an edge direction
                                                   (IN, OUT, NONE) */
      /* The following constant stores the 3-characters tag that identifies
         policy graphs when being written to file */
      static const char policyGraphMarker[4];

   public:
      PolicyGraph();                    // Default constructor
      ~PolicyGraph();                   // Destructor
      bool enhanced;                    /* If this flag is true, then
                                           this graph has been enhanced with
                                           "real" routing information */
      bool writeToFile(FILE*);          /* Dumps all the policy graph contents to an
                                           ASCII file; use `readFromFile' to
                                           get them back; the only parameter is
                                           a pointer to an open file strem;
                                           this method returns true if operation
                                           is successful; false otherwise */
      bool readFromFile(FILE*);         /* Builds a policy graph from the data that have
                                           previously been generated by `writeToFile';
                                           the only parameter is a pointer to an
                                           open file stream; returns true if
                                           operation is successful; false otherwise */
      char *errReason();                /* Returns a string that describes the
                                           reason of last file error; this method
                                           should be invoked after that either one
                                           of `writeToFile' or `readFromFile'
                                           returned false */
};

// ************** BGP Policy graph *************

#endif
